import logging

import numpy as np

from pymchelper.detector import MeshAxis
from pymchelper.shieldhit.detector.detector_type import SHDetType
from pymchelper.shieldhit.detector.estimator_type import SHGeoType
from pymchelper.flair.Data import Usrbin, UsrTrack, unpackArray

logger = logging.getLogger(__name__)


class FlukaBinaryReader:
    def __init__(self, filename):
        self.filename = filename

    def read(self, detector, nscale=1):

        try:
            usr = Usrbin(self.filename)
            data = usr.readData(0)
            fdata = unpackArray(data)
        except IOError:
            usr = UsrTrack(self.filename)
            data = usr.readData(0)
            fdata = unpackArray(data)[:usr.detector[0].ne]
        usr.say()  # file,title,time,weight,ncase,nbatch
        for i, _ in enumerate(usr.detector):
            logger.debug("-" * 20 + (" Detector number %i " % i) + "-" * 20)
            usr.say(i)  # details for each detector

        # TODO read detector type
        detector.det = "FLUKA"

        # TODO read particle type
        detector.particle = 0

        # TODO read geo type
        detector.geotyp = SHGeoType.unknown

        # TODO cross-check statistics
        detector.nstat = usr.ncase

        # TODO figure out when more detectors are used
        nx = usr.detector[0].nx
        ny = usr.detector[0].ny
        nz = usr.detector[0].nz

        xmin = usr.detector[0].xlow
        ymin = usr.detector[0].ylow
        zmin = usr.detector[0].zlow

        xmax = usr.detector[0].xhigh
        ymax = usr.detector[0].yhigh
        zmax = usr.detector[0].zhigh

        detector.x = MeshAxis(n=nx, min_val=xmin, max_val=xmax,
                              name="X", unit="", binning=MeshAxis.BinningType.linear)
        detector.y = MeshAxis(n=ny, min_val=ymin, max_val=ymax,
                              name="Y", unit="", binning=MeshAxis.BinningType.linear)
        detector.z = MeshAxis(n=nz, min_val=zmin, max_val=zmax,
                              name="Z", unit="", binning=MeshAxis.BinningType.linear)

        detector.unit, detector.name = "", ""

        # TODO read detector type
        detector.dettyp = SHDetType.unknown

        detector.data_raw = np.array(fdata)
        if nscale != 1:
            detector.data *= nscale
            # 1 gigaelectron volt / gram = 1.60217662 x 10-7 Gy
            detector.data *= 1.60217662e-7

        detector.title = usr.detector[0].name.decode('ascii')



import logging

import numpy as np

from pymchelper.shieldhit.detector.detector_type import SHDetType
from pymchelper.shieldhit.detector.estimator_type import SHGeoType

logger = logging.getLogger(__name__)


class SHBinaryWriter:
    def __init__(self, filename, options):
        self.filename = filename

    def write(self, detector):
        # TODO
        pass


class TxtWriter:
    @staticmethod
    def _axis_name(geo_type, axis_no):
        cyl = ['R', 'PHI', 'Z']
        msh = ['X', 'Y', 'Z']
        if geo_type in (SHGeoType.cyl, SHGeoType.dcyl):
            return cyl[axis_no]
        else:
            return msh[axis_no]

    def __init__(self, filename, options):
        if filename.endswith(".txt"):
            self.filename = filename
        else:
            self.filename = filename + ".txt"
        self.ax = ''
        self.ay = ''
        self.az = ''

    @staticmethod
    def _header_first_line(det):
        """first line with detector geo type"""
        result = "#   DETECTOR OUTPUT\n"
        if det.geotyp in (SHGeoType.plane, SHGeoType.dplane, ):
            result = "#             DETECTOR OUTPUT PLANE/DPLANE\n"
        elif det.geotyp in (SHGeoType.zone, SHGeoType.dzone, ):
            result = "#             DETECTOR OUTPUT ZONE/DZONE\n"
        elif det.geotyp in (SHGeoType.msh, SHGeoType.dmsh, ):
            result = "#   DETECTOR OUTPUT MSH/DMSH\n"
        elif det.geotyp == SHGeoType.geomap:
            result = "#   DETECTOR OUTPUT GEOMAP\n"
        return result

    def _header_geometric_info(self, det):
        """next block - scoring object geometrical information"""

        from pymchelper.fortranformatter import format_d
        result = ""
        if det.geotyp in {SHGeoType.plane, SHGeoType.dplane}:
            result += "#   PLANE point(X,Y,Z)         :"
            result += "{:s}".format(format_d(10, 3, det.sx))
            result += "{:s}".format(format_d(10, 3, det.sy))
            result += "{:s}\n".format(format_d(10, 3, det.sz))
            result += "#   PLANE normal vect(Vx,Vy,Vz):"
            result += "{:s}".format(format_d(10, 3, det.nx))
            result += "{:s}".format(format_d(10, 3, det.ny))
            result += "{:s}\n".format(format_d(10, 3, det.nz))
        elif det.geotyp in {SHGeoType.zone, SHGeoType.dzone}:
            result += "#   ZONE START:{:6d} ZONE END:{:6d}\n".format(int(det.x.min_val), int(det.x.max_val))
        else:
            result += "#   {:s} BIN:{:6d} {:s} BIN:{:6d} {:s} BIN:{:6d}\n".format(self.ax, det.x.n,
                                                                                  self.ay, det.y.n,
                                                                                  self.az, det.z.n)
        return result

    @staticmethod
    def _header_scored_value(det):
        """scored value and optionally particle type"""
        result = ""
        if det.geotyp != SHGeoType.geomap:
            result += "#   JPART:{:6d} DETECTOR TYPE: {:s}\n".format(det.particle, str(det.dettyp).ljust(10))
        else:
            det_type_name = str(det.dettyp)
            if det.dettyp in (SHDetType.zone, SHDetType.medium, ):
                det_type_name += "#"
            result += "#                DETECTOR TYPE: {:s}\n".format(str(det_type_name).ljust(10))
        return result

    def _header_no_of_bins_and_prim(self, det):
        from pymchelper.fortranformatter import format_d

        header = ""
        # number of bins in each dimensions
        if det.geotyp not in (SHGeoType.plane, SHGeoType.dplane, SHGeoType.zone, SHGeoType.dzone):
            header += "#   {:s} START:{:s}".format(self.ax, format_d(10, 3, det.x.min_val))
            header += " {:s} START:{:s}".format(self.ay, format_d(10, 3, det.y.min_val))
            header += " {:s} START:{:s}\n".format(self.az, format_d(10, 3, det.z.min_val))
            header += "#   {:s} END  :{:s}".format(self.ax, format_d(10, 3, det.x.max_val))
            header += " {:s} END  :{:s}".format(self.ay, format_d(10, 3, det.y.max_val))
            header += " {:s} END  :{:s}\n".format(self.az, format_d(10, 3, det.z.max_val))

        # number of primaries
        header += "#   PRIMARIES:" + format_d(10, 3, det.nstat) + "\n"

        return header

    def write(self, det):
        from pymchelper.fortranformatter import format_e

        self.ax = self._axis_name(det.geotyp, 0)
        self.ay = self._axis_name(det.geotyp, 1)
        self.az = self._axis_name(det.geotyp, 2)

        header = self._header_first_line(det)

        header += self._header_geometric_info(det)

        header += self._header_scored_value(det)

        header += self._header_no_of_bins_and_prim(det)

        # original bdo2txt is not saving header data for some of cylindrical scorers, hence we do the same
        if det.geotyp in (SHGeoType.cyl, SHGeoType.dcyl, ) and \
                det.dettyp in (SHDetType.fluence, SHDetType.avg_energy, SHDetType.avg_beta, SHDetType.energy):
            header = ""

        # dump data
        with open(self.filename, 'w') as fout:
            logger.info("Writing: " + self.filename)
            fout.write(header)

            det_error = det.error_raw.ravel()
            if np.all(np.isnan(det.error_raw)):
                det_error = [None] * det.data_raw.size
            zlist, ylist, xlist = np.meshgrid(det.z.data, det.y.data, det.x.data, indexing='ij')
            for x, y, z, v, e in zip(xlist.ravel(), ylist.ravel(), zlist.ravel(), det.data.ravel(), det_error):
                if det.geotyp in {SHGeoType.zone, SHGeoType.dzone}:
                    x = 0.0
                # dirty hack to be compliant with old bdo2txt and files generated in old (<0.6) BDO format
                # this hack will be removed at some point together with bdo-style converter
                elif not hasattr(det, "mc_code_version") and det.geotyp == SHGeoType.plane:
                    x = (det.sx + det.nx) / 2.0
                    y = (det.sy + det.ny) / 2.0
                    z = (det.sz + det.nz) / 2.0
                else:
                    x = float('nan') if np.isnan(x) else x
                y = float('nan') if np.isnan(y) else y
                z = float('nan') if np.isnan(z) else z
                v = float('nan') if np.isnan(v) else v
                tmp = format_e(14, 7, x) + " " + format_e(14, 7, y) + " " + \
                    format_e(14, 7, z) + " " + format_e(23, 16, v)

                if e is not None:
                    e = float('nan') if np.isnan(e) else e
                    tmp += " " + format_e(23, 16, e)

                tmp += "\n"

                fout.write(tmp)



import logging

import numpy as np

logger = logging.getLogger(__name__)


class SparseWriter:
    """
    Supports writing sparse matrix format
    """
    def __init__(self, filename, options):
        self.filename = filename
        if not self.filename.endswith(".npz"):
            self.filename += ".npz"

        self.threshold = options.threshold
        logger.info("Sparse threshold {:g}".format(self.threshold))

    def write(self, detector):

        # detector.data array is a 3-D numpy array
        # some of its dimensions may be as well ones and the array reduced to 0,1 or 2-D
        all_items = detector.data.size
        logger.info("Number of all items: {:d}".format(all_items))

        # prepare a cut to select values which norm is greater than threshold
        # default value of threshold is zero, in this case non-zero values will be selected
        # cut will be 3-D arrays of booleans
        # note that numpy allocates here same amount of memory as for original data
        thres_cut = np.abs(detector.data) > self.threshold
        passed_items = np.sum(thres_cut)
        logger.info("Number of items passing threshold: {:d}".format(passed_items))
        logger.info("Sparse matrix compression rate: {:g}".format(passed_items / all_items))

        # select indices which pass threshold
        # we get here a plain python tuple of 3-elements
        # first element is numpy array of indices along X-axis, second for Y axis and third for Z
        # note that such table cannot be used directly to index numpy arrays
        indices = np.argwhere(thres_cut)

        # select data which pass threshold and save it as plain 1-D numpy array
        filtered_data = detector.data[thres_cut]

        # save file to NPZ file format
        np.savez(file=self.filename,
                 data=filtered_data,
                 indices=indices,
                 shape=detector.data.shape)



from collections import defaultdict
from enum import IntEnum
import logging
import os

from pymchelper.readers.fluka import FlukaBinaryReader
from pymchelper.readers.shieldhit import SHBinaryReader, SHTextReader


logger = logging.getLogger(__name__)


class Readers(IntEnum):
    """
    Available converters
    """
    fluka_bin = 0
    shieldhit_bin = 1
    fluka_txt = 2  # TODO to be implemented
    shieldhit_txt = 3

    @property
    @classmethod
    def _readers_mapping(cls):
        return {
            cls.fluka_bin: FlukaBinaryReader,
            cls.shieldhit_bin: SHBinaryReader,
            cls.shieldhit_txt: SHTextReader
        }

    @classmethod
    def fromname(cls, name):
        return cls._readers_mapping[Readers[name]]

    @classmethod
    def fromnumber(cls, number):
        return cls._readers_mapping[Readers(number)]


def guess_reader(filename):
    reader = SHTextReader(filename)

    # check if binary file is generated with SHIELD-HIT12A version > 0.6
    #  (in that case it may or may not have .bdo extension)
    # this check will also pass is file is generated with older SHIELD-HIT12A version
    #  (in that case we rely on the file extension)
    if SHBinaryReader(filename).test_version_0p6() or filename.endswith((".bdo", ".bdox")):
        reader = SHBinaryReader(filename)
    # find better way to discover if file comes from Fluka
    elif "_fort" in filename:
        reader = FlukaBinaryReader(filename)

    return reader


def group_input_files(input_file_list):
    """
    Takes set of input file names, belonging to possibly different estimators.
    Input files are grouped according to the estimators and for each group
    merging is performed, as in @merge_list method.
    Output file name is automatically generated.
    :param input_file_list: list of input files
    :return: core_names_dict
    """
    core_names_dict = defaultdict(list)
    # keys - core_name, value - list of full paths to corresponding files

    # loop over input list of file paths
    for filepath in input_file_list:

        # extract basename (strip directory part) for inspection
        basename = os.path.basename(filepath)

        # SHIELD-HIT12A binary file encountered
        if SHBinaryReader(filepath).test_version_0p6() or filepath.endswith(('.bdo', '.bdox')):
            # we expect the basename to follow one of two conventions:
            #  - corenameABCD.bdo (where ABCD is 4-digit integer)
            #  - corename.bdo
            core_name = basename[:-4]  # assume no number in the basename
            if basename[-8:-4].isdigit() and len(basename[-8:-4]) == 4:  # check if number present
                core_name = basename[:-8]
            core_names_dict[core_name].append(filepath)
        elif "_fort." in filepath:  # Fluka binary file encountered
            core_name = filepath[-2:]
            core_names_dict[core_name].append(filepath)

    return core_names_dict



from enum import IntEnum


class SHDetType(IntEnum):
    unknown = 0
    energy = 1
    fluence = 2
    crossflu = 3
    letflu = 4
    dose = 5
    dlet = 6
    tlet = 7
    avg_energy = 8
    avg_beta = 9
    material = 10
    ddd = 12
    alanine = 13
    counter = 14
    pet = 15
    dletg = 16
    tletg = 17
    zone = 18
    medium = 19
    rho = 20
    q = 21
    flu_char = 22
    flu_neut = 23
    let = 120  # for differential scoring
    angle = 121  # for differential scoring
    dose_gy = 205
    alanine_gy = 213

    def __str__(self):
        return self.name.upper().replace('_', '-')



#!/bin/env python
# $Id: Data.py 3608 2015-10-27 11:18:36Z bnv $
#
# Copyright and User License
# ~~~~~~~~~~~~~~~~~~~~~~~~~~
# Copyright Vasilis.Vlachoudis@cern.ch for the
# European Organization for Nuclear Research (CERN)
#
# All rights not expressly granted under this license are reserved.
#
# Installation, use, reproduction, display of the
# software ("flair"), in source and binary forms, are
# permitted free of charge on a non-exclusive basis for
# internal scientific, non-commercial and non-weapon-related
# use by non-profit organizations only.
#
# For commercial use of the software, please contact the main
# author Vasilis.Vlachoudis@cern.ch for further information.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the
#    distribution.
#
# DISCLAIMER
# ~~~~~~~~~~
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY, OF
# SATISFACTORY QUALITY, AND FITNESS FOR A PARTICULAR PURPOSE
# OR USE ARE DISCLAIMED. THE COPYRIGHT HOLDERS AND THE
# AUTHORS MAKE NO REPRESENTATION THAT THE SOFTWARE AND
# MODIFICATIONS THEREOF, WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADE SECRET OR OTHER PROPRIETARY RIGHT.
#
# LIMITATION OF LIABILITY
# ~~~~~~~~~~~~~~~~~~~~~~~
# THE COPYRIGHT HOLDERS AND THE AUTHORS SHALL HAVE NO
# LIABILITY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL,
# CONSEQUENTIAL, EXEMPLARY, OR PUNITIVE DAMAGES OF ANY
# CHARACTER INCLUDING, WITHOUT LIMITATION, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES, LOSS OF USE, DATA OR PROFITS,
# OR BUSINESS INTERRUPTION, HOWEVER CAUSED AND ON ANY THEORY
# OF CONTRACT, WARRANTY, TORT (INCLUDING NEGLIGENCE), PRODUCT
# LIABILITY OR OTHERWISE, ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGES.
#
# Author:	Vasilis.Vlachoudis@cern.ch
# Date:	24-Oct-2006

import re
import math
import struct

import pymchelper.flair.common.fortran as fortran
import pymchelper.flair.common.bmath as bmath
from pymchelper.flair.common.log import say

__author__ = "Vasilis Vlachoudis"
__email__ = "Vasilis.Vlachoudis@cern.ch"

_detectorPattern = re.compile(r"^ ?# ?Detector ?n?:\s*\d*\s*(.*)\s*", re.MULTILINE)
_blockPattern = re.compile(r"^ ?# ?Block ?n?:\s*\d*\s*(.*)\s*", re.MULTILINE)


# -------------------------------------------------------------------------------
# Unpack an array of floating point numbers
# -------------------------------------------------------------------------------
def unpackArray(data):
    return struct.unpack("=%df" % (len(data) // 4), data)


# ===============================================================================
# Empty class to fill with detector data
# ===============================================================================
class Detector:
    pass


# ===============================================================================
# Base class for all detectors
# ===============================================================================
class Usrxxx:
    def __init__(self, filename=None):
        """Initialize a USRxxx structure"""
        self.reset()
        if filename is None:
            return
        f = self.readHeader(filename)
        if f is not None and not f.closed:
            f.close()

    # ----------------------------------------------------------------------
    def reset(self):
        """Reset header information"""
        self.file = ""
        self.title = ""
        self.time = ""
        self.weight = 0
        self.ncase = 0
        self.nbatch = 0
        self.detector = []
        self.seekpos = -1
        self.statpos = -1

    # ----------------------------------------------------------------------
    # Read information from USRxxx file
    # @return the handle to the file opened
    # ----------------------------------------------------------------------
    def readHeader(self, filename):
        """Read header information, and return the file handle"""
        self.reset()
        self.file = filename
        f = open(self.file, "rb")

        # Read header
        data = fortran.read(f)
        if data is None:
            if not f.closed:
                f.close()
            raise IOError("Invalid USRxxx file")
        size = len(data)
        over1b = 0
        if size == 116:
            (title, time, self.weight) = \
                struct.unpack("=80s32sf", data)
            self.ncase = 1
            self.nbatch = 1
        elif size == 120:
            (title, time, self.weight, self.ncase) = \
                struct.unpack("=80s32sfi", data)
            self.nbatch = 1
        elif size == 124:
            (title, time, self.weight,
             self.ncase, self.nbatch) = \
                struct.unpack("=80s32sfii", data)
        elif size == 128:
            (title, time, self.weight,
             self.ncase, over1b, self.nbatch) = \
                struct.unpack("=80s32sfiii", data)
        else:
            if not f.closed:
                f.close()
            raise IOError("Invalid USRxxx file")

        if over1b > 0:
            self.ncase = int(self.ncase) + int(over1b) * 1000000000

        self.title = title.strip()
        self.time = time.strip()

        return f

    # ----------------------------------------------------------------------
    # Read detector data
    # ----------------------------------------------------------------------
    def readData(self, det):
        """Read detector det data structure"""
        f = open(self.file, "rb")
        fortran.skip(f)  # Skip header
        for i in range(2 * det):
            fortran.skip(f)  # Detector Header & Data
        fortran.skip(f)  # Detector Header
        data = fortran.read(f)
        f.close()
        return data

    # ----------------------------------------------------------------------
    # Read detector statistical data
    # ----------------------------------------------------------------------
    def readStat(self, det):
        """Read detector det statistical data"""
        if self.statpos < 0:
            return None
        f = open(self.file, "rb")
        f.seek(self.statpos)
        for _ in range(det):
            fortran.skip(f)  # Detector Data
        data = fortran.read(f)
        f.close()
        return data

    # ----------------------------------------------------------------------
    def sayHeader(self):
        say("File   : ", self.file)
        say("Title  : ", self.title)
        say("Time   : ", self.time)
        say("Weight : ", self.weight)
        say("NCase  : ", self.ncase)
        say("NBatch : ", self.nbatch)


# ===============================================================================
# Residual nuclei detector
# ===============================================================================
class Resnuclei(Usrxxx):
    # ----------------------------------------------------------------------
    # Read information from a RESNUCLEi file
    # Fill the self.detector structure
    # ----------------------------------------------------------------------
    def readHeader(self, filename):
        """Read residual nuclei detector information"""
        f = Usrxxx.readHeader(self, filename)
        self.nisomers = 0
        if self.ncase <= 0:
            self.evol = True
            self.ncase = -self.ncase

            data = fortran.read(f)
            nir = (len(data) - 4) // 8
            self.irrdt = struct.unpack("=i%df" % (2 * nir), data)
        else:
            self.evol = False
            self.irrdt = None

        for _ in range(1000):
            # Header
            data = fortran.read(f)
            if data is None:
                break
            size = len(data)
            self.irrdt = None

            # Statistics are present?
            if size == 14 and data[:8] == "ISOMERS:":
                self.nisomers = struct.unpack("=10xi", data)[0]
                data = fortran.read(f)
                data = fortran.read(f)
                size = len(data)

            if size == 14 and data[:10] == "STATISTICS":
                self.statpos = f.tell()
                break

            if size != 38:
                if not f.closed:
                    f.close()
                raise IOError("Invalid RESNUCLEi file header size=%d" % (size))

            # Parse header
            header = struct.unpack("=i10siif3i", data)

            det = Detector()
            det.nb = header[0]
            det.name = header[1].strip()
            det.type = header[2]
            det.region = header[3]
            det.volume = header[4]
            det.mhigh = header[5]
            det.zhigh = header[6]
            det.nmzmin = header[7]

            self.detector.append(det)

            if self.evol:
                data = fortran.read(f)
                self.tdecay = struct.unpack("=f", data)
            else:
                self.tdecay = 0.0

            size = det.zhigh * det.mhigh * 4
            if size != fortran.skip(f):
                raise IOError("Invalid RESNUCLEi file")

        f.close()

    # ----------------------------------------------------------------------
    # Read detector data
    # ----------------------------------------------------------------------
    def readData(self, n):
        """Read detector det data structure"""
        f = open(self.file, "rb")
        fortran.skip(f)
        if self.evol:
            fortran.skip(f)

        for _ in range(n):
            fortran.skip(f)  # Detector Header & Data
            if self.evol:
                fortran.skip(f)  # TDecay
            fortran.skip(f)  # Detector data
            if self.nisomers:
                fortran.skip(f)  # Isomers header
                fortran.skip(f)  # Isomers data

        fortran.skip(f)  # Detector Header & Data
        if self.evol:
            fortran.skip(f)  # TDecay
        data = fortran.read(f)  # Detector data
        f.close()
        return data

    # ----------------------------------------------------------------------
    # Read detector statistical data
    # ----------------------------------------------------------------------
    def readStat(self, n):
        """Read detector det statistical data"""
        if self.statpos < 0:
            return None
        f = open(self.file, "rb")
        f.seek(self.statpos)

        f.seek(self.statpos)
        if self.nisomers:
            nskip = 7 * n
        else:
            nskip = 6 * n
        for i in range(nskip):
            fortran.skip(f)  # Detector Data

        total = fortran.read(f)
        A = fortran.read(f)
        errA = fortran.read(f)
        Z = fortran.read(f)
        errZ = fortran.read(f)
        data = fortran.read(f)
        if self.nisomers:
            iso = fortran.read(f)
        else:
            iso = None
        f.close()
        return (total, A, errA, Z, errZ, data, iso)

    # ----------------------------------------------------------------------
    def say(self, det=None):
        """print header/detector information"""
        if det is None:
            self.sayHeader()
        else:
            bin = self.detector[det]
            say("Bin    : ", bin.nb)
            say("Title  : ", bin.name)
            say("Type   : ", bin.type)
            say("Region : ", bin.region)
            say("Volume : ", bin.volume)
            say("Mhigh  : ", bin.mhigh)
            say("Zhigh  : ", bin.zhigh)
            say("NMZmin : ", bin.nmzmin)


# ===============================================================================
# Usrbdx Boundary Crossing detector
# ===============================================================================
class Usrbdx(Usrxxx):
    # ----------------------------------------------------------------------
    # Read information from a USRBDX file
    # Fill the self.detector structure
    # ----------------------------------------------------------------------
    def readHeader(self, filename):
        """Read boundary crossing detector information"""
        f = Usrxxx.readHeader(self, filename)

        for _ in range(1000):
            # Header
            data = fortran.read(f)
            if data is None:
                break
            size = len(data)

            # Statistics are present?
            if size == 14:
                # In statistics
                #   1: total, error
                #   2: N,NG,Elow (array with Emaxi)
                #   3: Differential integrated over solid angle
                #   4: -//- errors
                #   5: Cumulative integrated over solid angle
                #   6: -//- errors
                #   7: Double differential data
                self.statpos = f.tell()
                for det in self.detector:
                    data = unpackArray(fortran.read(f))
                    det.total = data[0]
                    det.totalerror = data[1]
                    for j in range(6):
                        fortran.skip(f)
                break
            if size != 78:
                if not f.closed:
                    f.close()
                raise IOError("Invalid USRBDX file")

            # Parse header

            header = struct.unpack("=i10siiiifiiiffifffif", data)

            det = Detector()
            det.nb = header[0]  # mx
            det.name = header[1].strip()  # titusx
            det.type = header[2]  # itusbx
            det.dist = header[3]  # idusbx
            det.reg1 = header[4]  # nr1usx
            det.reg2 = header[5]  # nr2usx
            det.area = header[6]  # ausbdx
            det.twoway = header[7]  # lwusbx
            det.fluence = header[8]  # lfusbx
            det.lowneu = header[9]  # llnusx
            det.elow = header[10]  # ebxlow
            det.ehigh = header[11]  # ebxhgh
            det.ne = header[12]  # nebxbn
            det.de = header[13]  # debxbn
            det.alow = header[14]  # abxlow
            det.ahigh = header[15]  # abxhgh
            det.na = header[16]  # nabxbn
            det.da = header[17]  # dabxbn

            self.detector.append(det)

            if det.lowneu:
                data = fortran.read(f)
                det.ngroup = struct.unpack("=i", data[:4])[0]
                det.egroup = struct.unpack("=%df" % (det.ngroup + 1), data[4:])
            else:
                det.ngroup = 0
                det.egroup = []

            size = (det.ngroup + det.ne) * det.na * 4
            if size != fortran.skip(f):
                raise IOError("Invalid USRBDX file")
        f.close()

    # ----------------------------------------------------------------------
    # Read detector data
    # ----------------------------------------------------------------------
    def readData(self, n):
        """Read detector n data structure"""
        f = open(self.file, "rb")
        fortran.skip(f)
        for i in range(n):
            fortran.skip(f)  # Detector Header
            if self.detector[i].lowneu:
                fortran.skip(f)  # Detector low enetry neutron groups
            fortran.skip(f)  # Detector data

        fortran.skip(f)  # Detector Header
        if self.detector[n].lowneu:
            fortran.skip(f)  # Detector low enetry neutron groups
        data = fortran.read(f)  # Detector data
        f.close()
        return data

    # ----------------------------------------------------------------------
    # Read detector statistical data
    # ----------------------------------------------------------------------
    def readStat(self, n):
        """Read detector n statistical data"""
        if self.statpos < 0:
            return None
        f = open(self.file, "rb")
        f.seek(self.statpos)
        for i in range(n):
            for j in range(7):
                fortran.skip(f)  # Detector Data

        for j in range(6):
            fortran.skip(f)  # Detector Data
        data = fortran.read(f)
        f.close()
        return data

    # ----------------------------------------------------------------------
    def say(self, det=None):
        """print header/detector information"""
        if det is None:
            self.sayHeader()
        else:
            det = self.detector[det]
            say("BDX    : ", det.nb)
            say("Title  : ", det.name)
            say("Type   : ", det.type)
            say("Dist   : ", det.dist)
            say("Reg1   : ", det.reg1)
            say("Reg2   : ", det.reg2)
            say("Area   : ", det.area)
            say("2way   : ", det.twoway)
            say("Fluence: ", det.fluence)
            say("LowNeu : ", det.lowneu)
            say("Energy : [", det.elow, "..", det.ehigh, "] ne=", det.ne, "de=", det.de)
            if det.lowneu:
                say("LOWNeut : [", det.egroup[-1], "..", det.egroup[0], "] ne=", det.ngroup)
            say("Angle  : [", det.alow, "..", det.ahigh, "] na=", det.na, "da=", det.da)
            say("Total  : ", det.total, "+/-", det.totalerror)


# ===============================================================================
# Usrtrack detector
# ===============================================================================
class UsrTrack(Usrxxx):
    # ----------------------------------------------------------------------
    # Read information from TRACK file
    # Fill the self.detector structure
    # ----------------------------------------------------------------------
    def readHeader(self, filename):
        """Read USRTRACK detector information"""
        f = Usrxxx.readHeader(self, filename)

        for _ in range(1000):
            # Header
            data = fortran.read(f)
            if data is None:
                break
            size = len(data)

            if size != 50:
                if not f.closed:
                    f.close()
                raise IOError("Invalid TRACK file")

            # Parse header
            # see http://www.fluka.org/flair/ustsuw.f for reference
            header = struct.unpack("=i10siiififfif", data)

            bin_det = Detector()
            bin_det.nb = header[0]  # mtc
            bin_det.name = header[1].strip()  # TITUTC
            bin_det.type = header[2]  # ITUSTC
            bin_det.region = header[4]  # IDUSTC
            bin_det.volume = header[5]  # VUSRTC
            bin_det.low_en_neutr_sc = header[6]  # LLNUTC
            bin_det.elow = header[7]  # ETCLOW minimum energy
            bin_det.ehigh = header[8]  # ETCHGH maximum energy
            bin_det.ne = header[9]  # NETCBN number of energy intervals
            bin_det.de = header[10]  # DETCBN energy bin width

            bin_det.xlow = bin_det.elow
            bin_det.xhigh = bin_det.ehigh
            bin_det.nx = header[6]
            bin_det.nx = bin_det.ne

            bin_det.ylow = 0.0
            bin_det.yhigh = 0.0
            bin_det.ny = 1

            bin_det.zlow = 0.0
            bin_det.zhigh = 0.0
            bin_det.nz = 1

            self.detector.append(bin_det)

            if bin_det.low_en_neutr_sc:
                data = fortran.read(f)
                bin_det.ngroup = struct.unpack("=i", data[:4])[0]
                bin_det.egroup = struct.unpack("=%df" % (bin_det.ngroup + 1), data[4:])
            else:
                bin_det.ngroup = 0
                bin_det.egroup = []

            size = (bin_det.ngroup + bin_det.ne) * 4
            if size != fortran.skip(f):
                raise IOError("Invalid USRTRACK file")
        f.close()

    # ----------------------------------------------------------------------
    # Read detector data
    # ----------------------------------------------------------------------
    def readData(self, n):
        """Read detector n data structure"""
        f = open(self.file, "rb")
        fortran.skip(f)
        for i in range(n):
            fortran.skip(f)  # Detector Header
            if self.detector[i].low_en_neutr_sc:
                fortran.skip(f)  # Detector low enetry neutron groups
            fortran.skip(f)  # Detector data

        fortran.skip(f)  # Detector Header
        if self.detector[n].low_en_neutr_sc:
            fortran.skip(f)  # Detector low enetry neutron groups
        data = fortran.read(f)  # Detector data
        f.close()
        return data

    # ----------------------------------------------------------------------
    # Read detector statistical data
    # ----------------------------------------------------------------------
    def readStat(self, n):
        """Read detector n statistical data"""
        if self.statpos < 0:
            return None
        f = open(self.file, "rb")
        f.seek(self.statpos)
        for i in range(n):
            fortran.skip(f)  # Detector Data
        data = fortran.read(f)
        f.close()
        return data

    # ----------------------------------------------------------------------
    def say(self, det=None):
        """print header/detector information"""
        if det is None:
            self.sayHeader()
        else:
            bin = self.detector[det]
            say("Bin    : ", bin.nb)
            say("Title  : ", bin.name)
            say("Type   : ", bin.type)
            say("E      : [", bin.elow, "-", bin.ehigh, "] x", bin.ne, "dx=", bin.de)


# ===============================================================================
# Usrbin detector
# ===============================================================================
class Usrbin(Usrxxx):
    # ----------------------------------------------------------------------
    # Read information from USRBIN file
    # Fill the self.detector structure
    # ----------------------------------------------------------------------
    def readHeader(self, filename):
        """Read USRBIN detector information"""
        f = Usrxxx.readHeader(self, filename)

        for _ in range(1000):
            # Header
            data = fortran.read(f)
            if data is None:
                break
            size = len(data)

            # Statistics are present?
            if size == 14 and data[:10] == "STATISTICS":
                self.statpos = f.tell()
                break
            if size != 86:
                if not f.closed:
                    f.close()
                raise IOError("Invalid USRBIN file")

            # Parse header
            header = struct.unpack("=i10siiffifffifffififff", data)

            bin_det = Detector()
            bin_det.nb = header[0]
            bin_det.name = header[1].strip()
            bin_det.type = header[2]
            bin_det.score = header[3]

            bin_det.xlow = float(bmath.format_number(header[4], 9, useD=False))
            bin_det.xhigh = float(bmath.format_number(header[5], 9, useD=False))
            bin_det.nx = header[6]
            if bin_det.nx > 0 and bin_det.type not in (2, 12, 8, 18):
                bin_det.dx = (bin_det.xhigh - bin_det.xlow) / float(bin_det.nx)
            else:
                bin_det.dx = float(bmath.format_number(header[7], 9, useD=False))

            if bin_det.type in (1, 11):
                bin_det.ylow = -math.pi
                bin_det.yhigh = math.pi
            else:
                bin_det.ylow = float(bmath.format_number(header[8], 9, useD=False))
                bin_det.yhigh = float(bmath.format_number(header[9], 9, useD=False))
            bin_det.ny = header[10]
            if bin_det.ny > 0 and bin_det.type not in (2, 12, 8, 18):
                bin_det.dy = (bin_det.yhigh - bin_det.ylow) / float(bin_det.ny)
            else:
                bin_det.dy = float(bmath.format_number(header[11], 9, useD=False))

            bin_det.zlow = float(bmath.format_number(header[12], 9, useD=False))
            bin_det.zhigh = float(bmath.format_number(header[13], 9, useD=False))
            bin_det.nz = header[14]
            if bin_det.nz > 0 and bin_det.type not in (2, 12):  # 8=special with z=real
                bin_det.dz = (bin_det.zhigh - bin_det.zlow) / float(bin_det.nz)
            else:
                bin_det.dz = float(bmath.format_number(header[15], 9, useD=False))

            bin_det.lntzer = header[16]
            bin_det.bk = header[17]
            bin_det.b2 = header[18]
            bin_det.tc = header[19]

            self.detector.append(bin_det)

            size = bin_det.nx * bin_det.ny * bin_det.nz * 4
            if fortran.skip(f) != size:
                raise IOError("Invalid USRBIN file")
        f.close()

    # ----------------------------------------------------------------------
    # Read detector data
    # ----------------------------------------------------------------------
    def readData(self, n):
        """Read detector det data structure"""
        f = open(self.file, "rb")
        fortran.skip(f)
        for i in range(n):
            fortran.skip(f)  # Detector Header
            fortran.skip(f)  # Detector data
        fortran.skip(f)  # Detector Header
        data = fortran.read(f)  # Detector data
        f.close()
        return data

    # ----------------------------------------------------------------------
    # Read detector statistical data
    # ----------------------------------------------------------------------
    def readStat(self, n):
        """Read detector n statistical data"""
        if self.statpos < 0:
            return None
        f = open(self.file, "rb")
        f.seek(self.statpos)
        for i in range(n):
            fortran.skip(f)  # Detector Data
        data = fortran.read(f)
        f.close()
        return data

    # ----------------------------------------------------------------------
    def say(self, det=None):
        """print header/detector information"""
        if det is None:
            self.sayHeader()
        else:
            bin = self.detector[det]
            say("Bin    : ", bin.nb)
            say("Title  : ", bin.name)
            say("Type   : ", bin.type)
            say("Score  : ", bin.score)
            say("X      : [", bin.xlow, "-", bin.xhigh, "] x", bin.nx, "dx=", bin.dx)
            say("Y      : [", bin.ylow, "-", bin.yhigh, "] x", bin.ny, "dy=", bin.dy)
            say("Z      : [", bin.zlow, "-", bin.zhigh, "] x", bin.nz, "dz=", bin.dz)
            say("L      : ", bin.lntzer)
            say("bk     : ", bin.bk)
            say("b2     : ", bin.b2)
            say("tc     : ", bin.tc)


# ===============================================================================
# MGDRAW output
# ===============================================================================
class Mgdraw:
    def __init__(self, filename=None):
        """Initialize a MGDRAW structure"""
        self.reset()
        if filename is None:
            return
        self.open(filename)

    # ----------------------------------------------------------------------
    def reset(self):
        """Reset information"""
        self.file = ""
        self.hnd = None
        self.nevent = 0
        self.data = None

    # ----------------------------------------------------------------------
    # Open file and return handle
    # ----------------------------------------------------------------------
    def open(self, filename):
        """Read header information, and return the file handle"""
        self.reset()
        self.file = filename
        try:
            self.hnd = open(self.file, "rb")
        except IOError:
            self.hnd = None
        return self.hnd

    # ----------------------------------------------------------------------
    def close(self):
        self.hnd.close()

    # ----------------------------------------------------------------------
    # Read or skip next event from mgread structure
    # ----------------------------------------------------------------------
    def readEvent(self, e_type=None):
        # Read header
        data = fortran.read(self.hnd)
        if data is None:
            return None
        if len(data) == 20:
            ndum, mdum, jdum, edum, wdum \
                = struct.unpack("=iiiff", data)
        else:
            raise IOError("Invalid MGREAD file")

        self.nevent += 1

        if ndum > 0:
            if e_type is None or e_type == 0:
                self.readTracking(ndum, mdum, jdum, edum, wdum)
            else:
                fortran.skip(self.hnd)
            return 0
        elif ndum == 0:
            if e_type is None or e_type == 1:
                self.readEnergy(mdum, jdum, edum, wdum)
            else:
                fortran.skip(self.hnd)
            return 1
        else:
            if e_type is None or e_type == 2:
                self.readSource(-ndum, mdum, jdum, edum, wdum)
            else:
                fortran.skip(self.hnd)
            return 2

    # ----------------------------------------------------------------------
    def readTracking(self, ntrack, mtrack, jtrack, etrack, wtrack):
        self.ntrack = ntrack
        self.mtrack = mtrack
        self.jtrack = jtrack
        self.etrack = etrack
        self.wtrack = wtrack
        data = fortran.read(self.hnd)
        if data is None:
            raise IOError("Invalid track event")
        fmt = "=%df" % (3 * (ntrack + 1) + mtrack + 1)
        self.data = struct.unpack(fmt, data)
        return ntrack

    # ----------------------------------------------------------------------
    def readEnergy(self, icode, jtrack, etrack, wtrack):
        self.icode = icode
        self.jtrack = jtrack
        self.etrack = etrack
        self.wtrack = wtrack
        data = fortran.read(self.hnd)
        if data is None:
            raise IOError("Invalid energy deposition event")
        self.data = struct.unpack("=4f", data)
        return icode

    # ----------------------------------------------------------------------
    def readSource(self, ncase, npflka, nstmax, tkesum, weipri):
        self.ncase = ncase
        self.npflka = npflka
        self.nstmax = nstmax
        self.tkesum = tkesum
        self.weipri = weipri

        data = fortran.read(self.hnd)
        if data is None:
            raise IOError("Invalid source event")
        fmt = "=" + ("i8f" * npflka)
        self.data = struct.unpack(fmt, data)
        return ncase


# ===============================================================================
if __name__ == "__main__":
    import sys

    say("=" * 80)
    usr = Usrbdx(sys.argv[1])
    usr.say()
    for i, _ in enumerate(usr.detector):
        say("-" * 50)
        usr.say(i)
        data = unpackArray(usr.readData(i))
        stat = unpackArray(usr.readStat(i))



import logging

from pymchelper.shieldhit.detector.detector_type import SHDetType
from pymchelper.shieldhit.detector.estimator_type import SHGeoType
from pymchelper.shieldhit.detector.geometry import Geometry
from pymchelper.shieldhit.particle import SHParticleType, SHHeavyIonType

logger = logging.getLogger(__name__)


class SHEstimator:
    allowed_detectors = {
        SHGeoType.unknown: (),
        SHGeoType.zone: (SHDetType(i + 1) for i in range(SHDetType.flu_neut - 1)),
        SHGeoType.cyl: (SHDetType(i + 1) for i in range(SHDetType.flu_neut - 1)),
        SHGeoType.msh: (SHDetType(i + 1) for i in range(SHDetType.flu_neut - 1)),
        SHGeoType.plane: (SHDetType(i + 1) for i in range(SHDetType.flu_neut - 1)),
        SHGeoType.dzone: (SHDetType(i + 1) for i in range(SHDetType.flu_neut - 1)),
        SHGeoType.dcyl: (SHDetType(i + 1) for i in range(SHDetType.flu_neut - 1)),
        SHGeoType.dmsh: (SHDetType(i + 1) for i in range(SHDetType.flu_neut - 1)),
        SHGeoType.dplane: (SHDetType(i + 1) for i in range(SHDetType.flu_neut - 1)),
        SHGeoType.dcylz: (SHDetType(i + 1) for i in range(SHDetType.flu_neut - 1)),
        SHGeoType.dmshz: (SHDetType(i + 1) for i in range(SHDetType.flu_neut - 1)),
        SHGeoType.trace: (SHDetType.unknown, ),
        SHGeoType.voxscore: (SHDetType(i + 1) for i in range(SHDetType.flu_neut - 1)),
        SHGeoType.geomap: (SHDetType.zone, SHDetType.medium, SHDetType.rho),
    }

    def __init__(self):
        self.estimator = SHGeoType.unknown
        self.particle_type = SHParticleType.unknown
        self.heavy_ion_type = SHHeavyIonType()
        self.detector_type = SHDetType.unknown
        self.geometry = Geometry()
        self.filename = ""

    def is_valid(self):
        if self.estimator not in self.geometry.allowed_estimators():
            logger.error("{:s} is not a valid estimator for {:s} " "geometry".format(self.estimator, self.geometry))
            return False

        return True



import logging
from collections import namedtuple
from enum import IntEnum
import numpy as np
from distutils.version import LooseVersion

from pymchelper.detector import MeshAxis
from pymchelper.shieldhit.detector.detector_type import SHDetType
from pymchelper.shieldhit.detector.estimator_type import SHGeoType

logger = logging.getLogger(__name__)


def _get_mesh_units(detector, axis):
    """ Set units depending on detector type.
    """

    _geotyp_units = {
        SHGeoType.msh: ("cm", "cm", "cm"),
        SHGeoType.dmsh: ("cm", "cm", "cm"),
        SHGeoType.cyl: ("cm", "radians", "cm"),
        SHGeoType.dcyl: ("cm", "radians", "cm"),
        SHGeoType.zone: ("zone number", "(nil)", "(nil)"),
        SHGeoType.voxscore: ("cm", "cm", "cm"),
        SHGeoType.geomap: ("cm", "cm", "cm"),
        SHGeoType.plane: ("cm", "cm", "cm"),
        SHGeoType.dplane: ("cm", "cm", "cm")
    }
    _default_units = ("(nil)", "(nil)", "(nil)")

    unit = _geotyp_units.get(detector.geotyp, _default_units)[axis]

    if detector.geotyp in {SHGeoType.msh, SHGeoType.dmsh, SHGeoType.voxscore, SHGeoType.geomap,
                           SHGeoType.plane, SHGeoType.dplane}:
        name = ("Position (X)", "Position (Y)", "Position (Z)")[axis]
    elif detector.geotyp in {SHGeoType.cyl, SHGeoType.dcyl}:
        name = ("Radius (R)", "Angle (PHI)", "Position (Z)")[axis]
    else:
        name = ""

    # dirty hack to change the units for differential scorers
    if hasattr(detector, 'dif_axis') and hasattr(detector, 'dif_type') and axis == detector.dif_axis:
        if detector.dif_type == 1:
            unit, name = _get_detector_unit(SHDetType.energy, detector.geotyp)
        elif detector.dif_type == 2:
            unit, name = _get_detector_unit(SHDetType.let, detector.geotyp)
        elif detector.dif_type == 3:
            unit, name = _get_detector_unit(SHDetType.angle, detector.geotyp)
        else:
            unit, name = "", ""

    return unit, name


def _get_detector_unit(detector_type, geotyp):
    """
    TODO
    :param detector_type:
    :param geotyp:
    :return:
    """
    if geotyp == SHGeoType.zone:
        dose_units = ("MeV/primary", "Dose*volume")
        dose_gy_units = ("J", "Dose*volume")
        alanine_units = ("MeV/primary", "Alanine RE*Dose*volume")
        alanine_gy_units = ("J", "Alanine RE*Dose*volume")
    else:
        dose_units = (" MeV/g/primary", "Dose")
        dose_gy_units = ("Gy", "Dose")
        alanine_units = ("MeV/g/primary", "Alanine RE*Dose")
        alanine_gy_units = ("Gy", "Alanine RE*Dose")

    _detector_units = {
        SHDetType.unknown: ("(nil)", "None"),
        SHDetType.energy: ("MeV/primary", "Energy"),
        SHDetType.fluence: ("cm^-2/primary", "Fluence"),
        SHDetType.crossflu: ("cm^-2/primary", "Planar fluence"),
        SHDetType.letflu: ("MeV/cm", "LET fluence"),
        SHDetType.dose: dose_units,
        SHDetType.dose_gy: dose_gy_units,
        SHDetType.dlet: ("keV/um", "dose-averaged LET"),
        SHDetType.tlet: ("keV/um", "track-averaged LET"),
        SHDetType.avg_energy: ("MeV", "Average energy"),
        SHDetType.avg_beta: ("(dimensionless)", "Average beta"),
        SHDetType.material: ("(nil)", "Material number"),
        SHDetType.alanine: alanine_units,
        SHDetType.alanine_gy: alanine_gy_units,
        SHDetType.counter: ("/primary", "Particle counter"),
        SHDetType.pet: ("/primary", "PET isotopes"),
        SHDetType.dletg: ("keV/um", "dose-averaged LET"),
        SHDetType.tletg: ("keV/um", "track-averaged LET"),
        SHDetType.q: ("(nil)", "beam quality Q"),
        SHDetType.flu_char: ("cm^-2/primary", "Charged particle fluence"),
        SHDetType.flu_neut: ("cm^-2/primary", "Neutral particle fluence"),
        SHDetType.let: ("keV/um", "LET"),
        SHDetType.angle: ("radians", "Angle"),
        SHDetType.zone: ("(dimensionless)", "Zone#"),
        SHDetType.medium: ("(dimensionless)", "Medium#"),
        SHDetType.rho: ("g/cm^3", "Density")
    }
    return _detector_units.get(detector_type, ("(nil)", "(nil)"))


def _postprocess(detector, nscale):
    """normalize result if we need that."""
    if detector.dettyp not in (SHDetType.dlet, SHDetType.tlet,
                               SHDetType.letflu,
                               SHDetType.dletg, SHDetType.tletg,
                               SHDetType.avg_energy, SHDetType.avg_beta,
                               SHDetType.material,
                               SHDetType.q):
        if detector.nstat != 0:  # geotyp = GEOMAP will have 0 projectiles simulated
            detector.data_raw /= np.float64(detector.nstat)
            detector.error_raw /= np.float64(detector.nstat)

    if nscale != 1:
        # scale with number of particles given by user
        detector.data_raw *= np.float64(nscale)
        detector.error_raw *= np.float64(nscale)

        # rescaling with particle number means also unit change for some detectors
        # from per particle to Grey - this is why we override detector type
        if detector.dettyp == SHDetType.dose:
            detector.dettyp = SHDetType.dose_gy
        if detector.dettyp == SHDetType.alanine:
            detector.dettyp = SHDetType.alanine_gy
        # for the same reason as above we change units
        if detector.dettyp in (SHDetType.dose_gy, SHDetType.alanine_gy):
            # 1 megaelectron volt / gram = 1.60217662 x 10-10 Gy
            MeV_g = np.float64(1.60217662e-10)
            detector.data_raw *= MeV_g
            detector.error_raw *= MeV_g
            detector.unit, detector.name = _get_detector_unit(detector.dettyp, detector.geotyp)


class SHBDOTagID(IntEnum):
    """ List of Tag ID numbers. Must be synchronized with sh_detect.h in SH12A.
    """

    # Hex values are used for better recognition in binary files, should they be inspected by humans.
    # Group 0x0000 - 0x00FF : Miscellaneous info
    shversion = 0x00    # [char*] full version string of shield-hit12a
    shbuilddate = 0x01  # [char*] date of build
    filedate = 0x02     # [char*] bdo file creation date, RFC 2822 compliant
    user = 0x03         # [char *] optional login name
    host = 0x04         # [char *] optional host where this file was created

    # Group 0xCC00 - 0xCCFF : Configuration
    dele = 0xCC00
    demin = 0xCC01
    itypst = 0xCC02
    itypms = 0xCC03
    oln = 0xCC04
    inucre = 0xCC05
    iemtrans = 0xCC06
    iextspec = 0xCC07
    intrfast = 0xCC08
    intrslow = 0xCC09
    apzlscl = 0xCC0A
    ioffset = 0xCC0B
    irifimc = 0xCC0C
    irifitrans = 0xCC0D
    irifizone = 0xCC0E
    ext_nproj = 0xCC0F
    ext_ptvdose = 0xCC10
    ixfirs = 0xCC11

    # Group 0xCE00 - 0xCEFF : CT specific tags
    ct_ang = 0xCE00   # holds two doubles with the couch and gantry angle
    ct_icnt = 0xCE01  # holds three
    ct_len = 0xCE02   # holds three

    # Group 0xEE00 - 0xEEFF : Estimator specific tags
    # estimator specific tags
    est_geotyp = 0xEE00    # may differ from det_geotyp in case of spc
    est_pages = 0xEE01     # number of detectors / pages for this estimator

    # Group 0xDD00 - 0xDDFF : Detector/page specific tags
    det_geotyp = 0xDD00      # may differ from est_geotyp in case of spc
    det_nbin = 0xDD01        # idet(1-3) (len=3) number of bins x,y,z
    det_part = 0xDD02        # idet(4) particle type which was scored
    det_dtype = 0xDD03       # idet(5) detector type
    det_partz = 0xDD04       # idet(6)
    det_parta = 0xDD05       # idet(7)
    det_dmat = 0xDD06        # idet(8)
    det_nbine = 0xDD07       # idet(9) number of bins in diff scorer, negative means log binning
    det_difftype = 0xDD08    # idet(10) detector type for differential scorer (i.e. angle, energy, let)
    det_zonestart = 0xDD09   # idet(11)
    det_dsize = 0xDD0A       # idet(12)
    det_dsizexyz = 0xDD0B    # idet(13)

    det_xyz_start = 0xDD0C   # det(1-3)
    det_xyz_stop = 0xDD0D    # det(4-6)
    det_dif_start = 0xDD0E   # det(7)
    det_dif_stop = 0xDD0F    # det(8)
    det_voxvol = 0xDD10      # det(9)
    det_thresh = 0xDD11       # det(10) lower energy scoring threshold

    det_data = 0xDDBB        # data block

    # Group 0xAA00 - 0xAAFF : Runtime variables
    rt_nstat = 0xAA00        # number of actually simulated particles
    rt_time = 0xAA01         # [unsigned long int] optional runtime in seconds


# for future use
# mapping = {SHBDOTagID.shversion: "mc_code_version",
#           SHBDOTagID.filedate: "filedate",
#           SHBDOTagID.user: "user",
#           SHBDOTagID.host: "host",
#           SHBDOTagID.rt_nstat: "nstat",
#           SHBDOTagID.det_dtype: "dettyp",
#           SHBDOTagID.est_geotyp: "geotyp",
#           SHBDOTagID.det_xyz_start: ("xmin", "ymin", "zmin"),
#           SHBDOTagID.det_xyz_stop: ("xmax", "ymax", "zmax"),
#           SHBDOTagID.det_nbin: ("nx", "ny", "nz")}


class SHBinaryReader:
    """
    Reads binary output files generated by SHIELD-HIT12A code.
    """
    def __init__(self, filename):
        self.filename = filename

    def test_version_0p6(self):
        sh_bdo_magic_number = b'xSH12A'
        with open(self.filename, "rb") as f:
            d1 = np.dtype([('magic', 'S6'),
                           ('end', 'S2'),
                           ('vstr', 'S16')])
            x = np.fromfile(f, dtype=d1, count=1)

            # if magic string is present, deep check for version number
            if sh_bdo_magic_number == x['magic'][0]:
                ver = x['vstr'][0].decode('ASCII')
                return (sh_bdo_magic_number == x['magic'][0]) and (LooseVersion(ver) >= LooseVersion("0.6"))
            else:
                return False

    def read(self, detector, nscale=1):
        if self.test_version_0p6():
            reader = _SHBinaryReader0p6(self.filename)
            reader.read(detector)
        else:
            reader = _SHBinaryReader0p1(self.filename)
            reader.read_header(detector)
            reader.read_payload(detector)
        _postprocess(detector, nscale)


def _bintyp(n):
    """
    Calculates type of binning based on number of bins.

    We follow the convention that positive number of bins means linear binning,
    while the negative - logarithmic.

    :param n: number of bins
    :return: MeshAxis.BinningType.linear or MeshAxis.BinningType.logarithmic
    """
    return MeshAxis.BinningType.linear if n > 0 else MeshAxis.BinningType.logarithmic


class _SHBinaryReader0p6:
    """
    Binary format reader from version >= 0.6
    """
    def __init__(self, filename):
        self.filename = filename

    def read(self, detector):
        logger.info("Reading: " + self.filename)
        with open(self.filename, "rb") as f:
            d1 = np.dtype([('magic', 'S6'),
                           ('end', 'S2'),
                           ('vstr', 'S16')])

            _x = np.fromfile(f, dtype=d1, count=1)  # read the data into numpy
            logger.debug("Magic : " + _x['magic'][0].decode('ASCII'))
            logger.debug("Endian: " + _x['end'][0].decode('ASCII'))
            logger.debug("VerStr: " + _x['vstr'][0].decode('ASCII'))

            while f:
                token = self.get_token(f)
                if token is None:
                    break

                pl_id, _pl_type, _pl_len, _pl = token

                pl = [None] * _pl_len

                # decode all strings (currently there will never be more than one per token)
                if 'S' in _pl_type.decode('ASCII'):
                    for i, _j in enumerate(_pl):
                        pl[i] = _pl[i].decode('ASCII').strip()
                else:
                    pl = _pl

                logger.debug("Read token {:s} 0x{:02x}".format(_pl_type.decode('ASCII'), pl_id))

                # TODO: some clever mapping could be done here surely
                # something like this: however the keymaps are not complete
                # attr_keys = SHBDOTagID(pl_id)
                # attributes = mapping[attr_keys]
                # for i, attr in enumerate(attributes):
                #     setattr(detector,attr,pl[i])
                #

                if pl_id == SHBDOTagID.shversion:
                    detector.mc_code_version = pl[0]
                    logger.debug("MC code version:" + detector.mc_code_version)

                if pl_id == SHBDOTagID.filedate:
                    detector.filedate = pl[0]

                if pl_id == SHBDOTagID.user:
                    detector.user = pl[0]

                if pl_id == SHBDOTagID.host:
                    detector.host = pl[0]

                if pl_id == SHBDOTagID.rt_nstat:
                    detector.nstat = pl[0]

                    # estimator block here ---
                if pl_id == SHBDOTagID.est_geotyp:
                    detector.geotyp = SHGeoType[pl[0].strip().lower()]

                if pl_id == SHBDOTagID.ext_ptvdose:
                    detector.tripdose = 0.0

                if pl_id == SHBDOTagID.ext_nproj:
                    detector.tripntot = -1

                if pl_id == SHBDOTagID.est_pages:
                    detector.pages = pl[0]
                    # todo: handling of multiple detectors (SPC)

                # read a single detector
                if pl_id == SHBDOTagID.det_dtype:
                    detector.dettyp = SHDetType(pl[0])

                if pl_id == SHBDOTagID.det_part:  # particle to be scored
                    detector.particle = pl[0]
                if pl_id == SHBDOTagID.det_partz:  # particle to be scored
                    detector.particle_z = pl[0]
                if pl_id == SHBDOTagID.det_parta:  # particle to be scored
                    detector.particle_a = pl[0]

                if pl_id == SHBDOTagID.det_nbin:
                    nx = pl[0]
                    ny = pl[1]
                    nz = pl[2]

                if pl_id == SHBDOTagID.det_xyz_start:
                    xmin = pl[0]
                    ymin = pl[1]
                    zmin = pl[2]

                if pl_id == SHBDOTagID.det_xyz_stop:
                    xmax = pl[0]
                    ymax = pl[1]
                    zmax = pl[2]

                # partial support for differential scoring (only linear binning)
                # TODO add some support for DMSH, DCYL and DZONE
                # TODO add support for logarithmic binning
                diff_geotypes = {SHGeoType.dplane, SHGeoType.dmsh, SHGeoType.dcyl, SHGeoType.dzone}
                if hasattr(detector, 'geotyp') and detector.geotyp in diff_geotypes:
                    if pl_id == SHBDOTagID.det_dif_start:
                        detector.dif_min = pl[0]

                    if pl_id == SHBDOTagID.det_dif_stop:
                        detector.dif_max = pl[0]

                    if pl_id == SHBDOTagID.det_nbine:
                        detector.dif_n = pl[0]

                    if pl_id == SHBDOTagID.det_difftype:
                        detector.dif_type = pl[0]

                if pl_id == SHBDOTagID.det_zonestart:
                    detector.zone_start = pl[0]

                if pl_id == SHBDOTagID.det_data:
                    detector.data_raw = np.asarray(pl)

            # TODO: would be better to not overwrite x,y,z and make proper case for ZONE scoring later.
            if detector.geotyp in {SHGeoType.zone, SHGeoType.dzone}:
                # special case for zone scoring, x min and max will be zone numbers
                xmin = detector.zone_start
                xmax = xmin + nx - 1
                ymin = 0.0
                ymax = 0.0
                zmin = 0.0
                zmax = 0.0
            elif detector.geotyp in {SHGeoType.plane, SHGeoType.dplane}:
                # special case for plane scoring, according to documentation we have:
                #  xmin, ymin, zmin = Sx, Sy, Sz (point on the plane)
                #  xmax, ymax, zmax = nx, ny, nz (normal vector)
                # to avoid situation where i.e. xmax < xmin (corresponds to nx < Sx)
                # we store only point on the plane
                detector.sx, detector.sy, detector.sz = xmin, ymin, zmin
                detector.nx, detector.ny, detector.nz = xmax, ymax, zmax
                xmax = xmin
                ymax = ymin
                zmax = zmin

            # check if scoring quantity is LET, if yes, than change units from [MeV/cm] to [keV/um]
            if hasattr(detector, 'dif_type') and detector.dif_type == 2:
                detector.dif_min /= 10.0
                detector.dif_max /= 10.0

            # # differential scoring data replacement
            if hasattr(detector, 'dif_min') and hasattr(detector, 'dif_max') and hasattr(detector, 'dif_n'):
                if nz == 1:
                    # max two axis (X or Y) filled with scored value, Z axis empty
                    # we can put differential quantity as Z axis
                    nz = detector.dif_n
                    zmin = detector.dif_min
                    zmax = detector.dif_max
                    detector.dif_axis = 2
                elif ny == 1:
                    # Z axis filled with scored value (X axis maybe also), Y axis empty
                    # we can put differential quantity as Y axis
                    ny = detector.dif_n
                    ymin = detector.dif_min
                    ymax = detector.dif_max
                    detector.dif_axis = 1
                    # due to the fact that the data in BDO file is arranged differently
                    # than assumed here, we need to do some reshaping
                    #   - in the raw BDO file loop goes first over scored value, then over differential quantity
                    #   - here we have (X-scored, Y-differential, Z-scored), so we need to reshape data from file
                    # such reshaping is not needed in other cases - there diff. value goes always as last axis
                    # first we get the reshaped view (3-D matrix)
                    reshaped_view = detector.data_raw.reshape((nx, ny, nz))
                    # reshape the data structure by swapping data and differential columns
                    detector.data_raw = np.transpose(reshaped_view, axes=(0, 2, 1)).flatten()
                elif nx == 1:
                    nx = detector.dif_n
                    xmin = detector.dif_min
                    xmax = detector.dif_max
                    detector.dif_axis = 0

            xunit, xname = _get_mesh_units(detector, 0)
            yunit, yname = _get_mesh_units(detector, 1)
            zunit, zname = _get_mesh_units(detector, 2)

            detector.x = MeshAxis(n=np.abs(nx), min_val=xmin, max_val=xmax, name=xname, unit=xunit, binning=_bintyp(nx))
            detector.y = MeshAxis(n=np.abs(ny), min_val=ymin, max_val=ymax, name=yname, unit=yunit, binning=_bintyp(ny))
            detector.z = MeshAxis(n=np.abs(nz), min_val=zmin, max_val=zmax, name=zname, unit=zunit, binning=_bintyp(nz))

            detector.unit, detector.name = _get_detector_unit(detector.dettyp, detector.geotyp)

            logger.debug("Done reading bdo file.")
            logger.debug("Detector data : " + str(detector.data))
            logger.debug("Detector nstat: " + str(detector.nstat))
            logger.debug("Detector nx   : " + str(detector.x.n))
            logger.debug("Detector ny   : " + str(detector.y.n))
            logger.debug("Detector nz   : " + str(detector.z.n))
            detector.counter = 1

    def get_token(self, f):
        """
        returns a tuple with 4 elements:
        0: payload id
        1: payload dtype string
        2: payload number of elements
        3: payload itself
        f is an open and readable file pointer.
        returns None if no token was found / EOF
        """

        tag = np.dtype([('pl_id', '<u8'),
                        ('pl_type', 'S8'),
                        ('pl_len', '<u8')])

        x1 = np.fromfile(f, dtype=tag, count=1)  # read the data into numpy

        if not x1:
            return None
        else:
            pl_id = x1['pl_id'][0]
            pl_type = x1['pl_type'][0]
            pl_len = x1['pl_len'][0]
            pl = np.fromfile(f,
                             dtype=pl_type,
                             count=pl_len)  # read the data into numpy
            return (pl_id, pl_type, pl_len, pl)


class _SHBinaryReader0p1:
    """
    Binary format reader from 0.1 <= version <= 0.6
    """
    def __init__(self, filename):
        self.filename = filename

    def read_header(self, detector):
        logger.info("Reading header: " + self.filename)

        detector.tripdose = 0.0
        detector.tripntot = -1

        # effective read
        # first figure out if this is a VOXSCORE card
        header_dtype = np.dtype([('__fo1', '<i4'), ('geotyp', 'S10')])
        header = np.fromfile(self.filename, header_dtype, count=1)

        if 'VOXSCORE' in header['geotyp'][0].decode('ascii'):
            header_dtype = np.dtype([('__fo1', '<i4'),     # 0x00
                                     ('geotyp', 'S10'),    # 0x04
                                     ('__fo2', '<i4'),     # 0x0E
                                     ('__fo3', '<i4'),     # 0x12
                                     ('nstat', '<i4'),     # 0x16 : nstat
                                     ('__fo4', '<i4'),     # 0x1A
                                     ('__foo1', '<i4'),    # 0x1E
                                     ('tds', '<f4'),       # 0x22 : tripdose
                                     ('__foo2', '<i4'),    # 0x26
                                     ('__foo3', '<i4'),    # 0x2A
                                     ('tnt', '<i8'),       # 0x2E : tripntot
                                     ('__foo4', '<i4'),    # 0x36
                                     ('__fo5', '<i4'),     # 0x3A
                                     # DET has 8x float64
                                     ('det', ('<f8', 8)),  # 0x3E : DET
                                     ('__fo6', '<i4'),     # 0x7E
                                     ('__fo7', '<i4'),     # 0x82
                                     # IDET has 11x int32
                                     ('idet', '<i4', 11),  # 0x86 : IDET
                                     ('__fo8', '<i4'),     # 0xB2
                                     ('reclen', '<i4')])   # 0xB6
            # payload starts at 0xBA (186)
            detector.payload_offset = 186
        else:
            # first figure out the length.
            header_dtype = np.dtype([('__fo1', '<i4'),
                                     ('geotyp', 'S10'),
                                     ('__fo2', '<i4'),
                                     ('__fo3', '<i4'),
                                     ('nstat', '<i4'),
                                     ('__fo4', '<i4'),
                                     ('__fo5', '<i4'),
                                     # DET has 8x float64
                                     ('det', ('<f8', 8)),  # DET
                                     ('__fo6', '<i4'),
                                     ('__fo7', '<i4'),
                                     # IDET has 11x int32
                                     ('idet', '<i4', 11),  # IDET
                                     ('__fo8', '<i4'),
                                     ('reclen', '<i4')])
            # payload starts at 0x9E (158)
            detector.payload_offset = 158

        header = np.fromfile(self.filename, header_dtype, count=1)
        detector.rec_size = header['reclen'][0] // 8

        if 'VOXSCORE' in header['geotyp'][0].decode('ascii'):
            detector.tripdose = header['tds'][0]
            detector.tripntot = header['tnt'][0]

        # map 10-elements table to namedtuple, for easier access
        # here is description of IDET table, assuming fortran-style numbering
        # (arrays starting from 1)
        # IDET(1) : Number of bins in first dimension. x or r or zones
        # IDET(2) : Number of bins in snd dimension, y or theta
        # IDET(3) : Number of bins in thrd dimension, z
        # IDET(4) : Particle type requested for scoring
        # IDET(5) : Detector type (see INITDET)
        # IDET(6) : Z of particle to be scored
        # IDET(7) : A of particle to be scored (only integers here)
        # IDET(8) : Detector material parameter
        # IDET(9) : Number of energy/amu (or LET) differential bins,
        #            negative if log.
        # IDET(10): Type of differential scoring, either LET, E/amu
        #            or polar angle
        # IDET(11): Starting zone of scoring for zone scoring
        DetectorAttributes = namedtuple('DetectorAttributes',
                                        ['dim_1_bins', 'dim_2_bins',
                                         'dim_3_bins',
                                         'particle_type', 'det_type',
                                         'particle_z', 'particle_a',
                                         'det_material',
                                         'diff_bins_no', 'diff_scoring_type',
                                         'starting_zone'])
        det_attribs = DetectorAttributes(*header['idet'][0])

        nx = det_attribs.dim_1_bins
        ny = det_attribs.dim_2_bins
        nz = det_attribs.dim_3_bins

        # DET(1-3): start positions for x y z or r theta z
        # DET(4-6): stop positions for x y z or r theta z
        # DET(7)  : start differential grid
        # DET(8)  : stop differential grid
        detector.det = header['det']
        detector.particle = det_attribs.particle_type

        try:
            detector.geotyp = SHGeoType[header['geotyp'][0].decode('ascii').strip().lower()]
        except Exception:
            detector.geotyp = SHGeoType.unknown
        detector.nstat = header['nstat'][0]

        if detector.geotyp not in {SHGeoType.zone, SHGeoType.dzone}:
            xmin = header['det'][0][0]
            ymin = header['det'][0][1]
            zmin = header['det'][0][2]

            xmax = header['det'][0][3]
            ymax = header['det'][0][4]
            zmax = header['det'][0][5]
        else:
            # special case for zone scoring, x min and max will be zone numbers
            xmin = det_attribs.starting_zone
            xmax = xmin + nx - 1
            ymin = 0.0
            ymax = 0.0
            zmin = 0.0
            zmax = 0.0

        if detector.geotyp in {SHGeoType.plane, SHGeoType.dplane}:
                # special case for plane scoring, according to documentation we have:
                #  xmin, ymin, zmin = Sx, Sy, Sz (point on the plane)
                #  xmax, ymax, zmax = nx, ny, nz (normal vector)
                # to avoid situation where i.e. xmax < xmin (corresponds to nx < Sx)
                # we store only point on the plane
                detector.sx, detector.sy, detector.sz = xmin, ymin, zmin
                detector.nx, detector.ny, detector.nz = xmax, ymax, zmax
                xmax = xmin
                ymax = ymin
                zmax = zmin

        xunit, xname = _get_mesh_units(detector, 0)
        yunit, yname = _get_mesh_units(detector, 1)
        zunit, zname = _get_mesh_units(detector, 2)

        detector.x = MeshAxis(n=np.abs(nx), min_val=xmin, max_val=xmax, name=xname, unit=xunit, binning=_bintyp(nx))
        detector.y = MeshAxis(n=np.abs(ny), min_val=ymin, max_val=ymax, name=yname, unit=yunit, binning=_bintyp(ny))
        detector.z = MeshAxis(n=np.abs(nz), min_val=zmin, max_val=zmax, name=zname, unit=zunit, binning=_bintyp(nz))

        detector.dettyp = SHDetType(det_attribs.det_type)

        detector.unit, detector.name = _get_detector_unit(detector.dettyp, detector.geotyp)

    # TODO: we need an alternative list, in case things have been scaled with nscale, since then things
    # are not "/particle" anymore.
    def read_payload(self, detector):
        logger.info("Reading data: " + self.filename)

        if detector.geotyp == SHGeoType.unknown or detector.dettyp == SHDetType.unknown:
            logger.error("Unknown geotyp or dettyp")
            return

        # next read the data:
        offset_str = "S" + str(detector.payload_offset)
        record_dtype = np.dtype([('trash', offset_str),
                                 ('bin2', '<f8', detector.rec_size)])
        record = np.fromfile(self.filename, record_dtype, count=-1)
        # BIN(*)  : a large array holding results. Accessed using pointers.
        detector.data_raw = np.array(record['bin2'][:][0])
        detector.counter = 1

    def read(self, detector):
        self.read_header(detector)
        self.read_payload(detector)


class SHTextReader:
    """
    Reads plain text files with data saved by binary-to-ascii converter.
    """
    def __init__(self, filename):
        self.filename = filename

    def read_header(self, detector):
        raise NotImplementedError

    def read_payload(self, detector):
        raise NotImplementedError

    def read(self, detector):
        self.read_header(detector)
        self.read_payload(detector)



from glob import glob
import logging
import os

import numpy as np

from pymchelper.detector import Detector, average_with_nan, ErrorEstimate
from pymchelper.readers.common import guess_reader, group_input_files
from pymchelper.writers.common import Converters

logger = logging.getLogger(__name__)


def fromfile(filename):
    """Read a detector data from a binary file ```filename```"""

    reader = guess_reader(filename)
    detector = Detector()
    detector.counter = 1
    reader.read(detector)
    detector.error_raw = np.zeros_like(detector.data_raw)
    detector.error_raw *= np.nan
    return detector


def fromfilelist(input_file_list, error, nan):
    """

    :param input_file_list:
    :param error:
    :param nan:
    :return:
    """
    if not isinstance(input_file_list, list):  # probably a string instead of list
        input_file_list = [input_file_list]

    if nan:
        detector_list = [fromfile(filename) for filename in input_file_list]
        result = average_with_nan(detector_list, error)
    elif len(input_file_list) == 1:
        result = fromfile(input_file_list[0])
    else:
        result = fromfile(input_file_list[0])

        # allocate memory for accumulator in standard deviation calculation
        # not needed if user requested not to include errors
        if error != ErrorEstimate.none:
            m2 = np.zeros_like(result.data_raw)

        # loop over all files
        for n, filename in enumerate(input_file_list[1:], start=2):
            x = fromfile(filename).data_raw

            # Running variance algorithm based on algorithm by B. P. Welford,
            # presented in Donald Knuth's Art of Computer Programming, Vol 2, page 232, 3rd edition.
            # Can be found here: http://www.johndcook.com/blog/standard_deviation/
            # and https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Online_algorithm
            delta = x - result.data_raw               # delta = x - mean
            result.data_raw += delta / n              # mean += delta / n
            if error != ErrorEstimate.none:
                m2 += delta * (x - result.data_raw)   # M2 += delta * (x - mean)

        # unbiased sample variance is stored in `__M2 / (n - 1)`
        # unbiased sample standard deviation in classical algorithm is calculated as (sqrt(1/(n-1)sum(x-<x>)**2)
        # here it is calculated as square root of unbiased sample variance:
        if len(input_file_list) > 1 and error != ErrorEstimate.none:
            result.error_raw = np.sqrt(m2 / (len(input_file_list) - 1))

        # if user requested standard error then we calculate it as:
        # S = stderr = stddev / sqrt(N), or in other words,
        # S = s/sqrt(N) where S is the corrected standard deviation of the mean.
        if len(input_file_list) > 1 and error == ErrorEstimate.stderr:
            result.error_raw /= np.sqrt(len(input_file_list))  # np.sqrt() always returns np.float64

    result.counter = len(input_file_list)
    core_names_dict = group_input_files(input_file_list)
    if len(core_names_dict) == 1:
        result.corename = list(core_names_dict)[0]

    return result


def frompattern(pattern, error, nan, jobs=-1, verbose=0):
    """

    :param pattern:
    :param error:
    :param nan:
    :param jobs:
    :param verbose:
    :return:
    """
    list_of_matching_files = glob(pattern)

    core_names_dict = group_input_files(list_of_matching_files)

    def _get_detector(filelist, error, nan):
        detector = fromfilelist(filelist, error, nan)
        return detector

    # parallel execution of output file generation, using all CPU cores
    # see http://pythonhosted.org/joblib
    try:
        from joblib import Parallel, delayed
        logger.info("Parallel processing on {:d} jobs (-1 means all)".format(jobs))
        # options.verbose count the number of `-v` switches provided by user
        # joblib Parallel class expects the verbosity as a larger number (i.e. multiple of 10)
        worker = Parallel(n_jobs=jobs, verbose=verbose * 10)
        result = worker(
            delayed(_get_detector)(filelist, error, nan)
            for core_name, filelist in core_names_dict.items()
        )
    except (ImportError, SyntaxError):
        # single-cpu implementation, in case joblib library fails (i.e. Python 3.2)
        logger.info("Single CPU processing")
        result = [_get_detector(core_name, filelist)
                  for core_name, filelist in core_names_dict.items()]
        return result


def convertfromlist(filelist, error, nan, outputdir, converter_name, options, outputfile=None):
    """

    :param filelist:
    :param error:
    :param nan:
    :param outputdir:
    :param converter_name:
    :param options:
    :param outputfile:
    :return:
    """
    detector = fromfilelist(filelist, error, nan)
    if outputfile is not None:
        output_path = outputfile
    elif outputdir is None:
        output_path = detector.corename
    else:
        output_path = os.path.join(outputdir, detector.corename)
    tofile(detector, output_path, converter_name, options)
    return None


def convertfrompattern(pattern, outputdir, converter_name, options,
                       error=ErrorEstimate.stderr, nan=True, jobs=-1, verbose=0):
    """

    :param pattern:
    :param outputdir:
    :param converter_name:
    :param options:
    :param error:
    :param nan:
    :param jobs:
    :param verbose:
    :return:
    """
    list_of_matching_files = glob(pattern)

    core_names_dict = group_input_files(list_of_matching_files)

    # parallel execution of output file generation, using all CPU cores
    # see http://pythonhosted.org/joblib
    try:
        from joblib import Parallel, delayed
        logger.info("Parallel processing on {:d} jobs (-1 means all)".format(jobs))
        # options.verbose count the number of `-v` switches provided by user
        # joblib Parallel class expects the verbosity as a larger number (i.e. multiple of 10)
        worker = Parallel(n_jobs=jobs, verbose=verbose * 10)
        worker(
            delayed(convertfromlist)(filelist, error, nan, outputdir, converter_name, options)
            for core_name, filelist in core_names_dict.items()
        )
    except (ImportError, SyntaxError):
        # single-cpu implementation, in case joblib library fails (i.e. Python 3.2)
        logger.info("Single CPU processing")
        for core_name, filelist in core_names_dict.items():
            convertfromlist(filelist, error, nan, outputdir, converter_name, options)


def tofile(detector, filename, converter_name, options):
    """
    Save a detector data to a ``filename`` using converter defined by ``converter_name``
    :param detector:
    :param filename:
    :param converter_name:
    :param options:
    :return:
    """
    writer_cls = Converters.fromname(converter_name)
    writer = writer_cls(filename, options)
    logger.debug("Writing file with corename {:s}".format(filename))
    writer.write(detector)



import logging
import os
from enum import IntEnum

import numpy as np

logger = logging.getLogger(__name__)


class PlotAxis(IntEnum):
    x = 1
    y = 2
    z = 3


class PlotDataWriter:
    def __init__(self, filename, options):
        self.filename = filename
        if not self.filename.endswith(".dat"):
            self.filename += ".dat"

    def write(self, detector):
        logger.info("Writing: " + self.filename)

        data_raw = detector.data_raw
        error_raw = detector.error_raw

        # change units for LET from MeV/cm to keV/um if necessary
        # a copy of data table is made here
        from pymchelper.shieldhit.detector.detector_type import SHDetType
        if detector.dettyp in (SHDetType.dlet, SHDetType.dletg, SHDetType.tlet, SHDetType.tletg):
            data_raw = data_raw * np.float64(0.1)  # 1 MeV / cm = 0.1 keV / um
            if not np.all(np.isnan(error_raw)) and np.any(error_raw):
                error_raw = error_raw * np.float64(0.1)  # 1 MeV / cm = 0.1 keV / um

        # special case for 0-dim data
        if detector.dimension == 0:
            # save two numbers to the file
            if not np.all(np.isnan(error_raw)) and np.any(error_raw):
                np.savetxt(self.filename, [[detector.data_raw, detector.error_raw]], fmt="%g %g", delimiter=' ')
            else:  # save one number to the file
                np.savetxt(self.filename, [detector.data_raw], fmt="%g", delimiter=' ')
        else:
            # each axis may have different number of points, this is what we store here:
            axis_data_columns_1d = [detector.plot_axis(i).data for i in range(detector.dimension)]

            # now we calculate running index for each axis
            axis_data_columns_long = [np.meshgrid(*axis_data_columns_1d, indexing='ij')[i].ravel()
                                      for i in range(len(axis_data_columns_1d))]

            fmt = "%g" + " %g" * detector.dimension
            data_to_save = axis_data_columns_long + [data_raw]

            # if error information is present save it as additional column
            if not np.all(np.isnan(error_raw)) and np.any(error_raw):
                fmt += " %g"
                data_to_save += [error_raw]

            # transpose from rows to columns
            data_columns = np.transpose(data_to_save)

            # save space-delimited text file
            np.savetxt(self.filename, data_columns, fmt=fmt, delimiter=' ')


class GnuplotDataWriter:
    def __init__(self, filename, options):
        self.data_filename = filename
        self.script_filename = filename
        self.plot_filename = filename

        if not self.plot_filename.endswith(".png"):
            self.plot_filename += ".png"
        if not self.script_filename.endswith(".plot"):
            self.script_filename += ".plot"
        if not self.data_filename.endswith(".dat"):
            self.data_filename += ".dat"

        dirname = os.path.split(self.script_filename)[0]
        self.awk_script_filename = os.path.join(dirname, "addblanks.awk")

    _awk_2d_script_content = """/^[[:blank:]]*#/ {next} # ignore comments (lines starting with #)
NF < 3 {next} # ignore lines which don't have at least 3 columns
$2 != prev {printf \"\\n\"; prev=$2} # print blank line
{print} # print the line
    """

    _header = """set term png
set output \"{plot_filename}\"
set title \"{title}\"
set xlabel \"{xlabel}\"
set ylabel \"{ylabel}\"
"""

    _error_plot_command = "'./{data_filename}' u 1:(max($2-$3,0.0)):($2+$3) w filledcurves " \
                          "fs transparent solid 0.2 lc 3 title '1-sigma confidence', "

    _plotting_command = {
        1: """max(x,y) = (x > y) ? x : y
plot {error_plot} './{data_filename}' u 1:2 w l lt 1 lw 2 lc -1 title 'mean value'
        """,
        2: """set view map
splot \"<awk -f addblanks.awk '{data_filename}'\" u 1:2:3 with pm3d
"""
    }

    def write(self, detector):
        # skip plotting 0-D and 3-D data
        if detector.dimension in {0, 3}:
            return

        # set labels
        plot_x_axis = detector.plot_axis(0)
        xlabel = ImageWriter._make_label(plot_x_axis.unit, plot_x_axis.name)
        if detector.dimension == 1:
            ylabel = ImageWriter._make_label(detector.unit, detector.name)
        elif detector.dimension == 2:
            plot_y_axis = detector.plot_axis(1)
            ylabel = ImageWriter._make_label(plot_y_axis.unit, plot_y_axis.name)

            # for 2-D plots write additional awk script to convert data
            # as described in gnuplot faq: http://www.gnuplot.info/faq/faq.html#x1-320003.9
            with open(self.awk_script_filename, 'w') as script_file:
                logger.info("Writing: " + self.awk_script_filename)
                script_file.write(self._awk_2d_script_content)

        # save gnuplot script
        with open(self.script_filename, 'w') as script_file:
            logger.info("Writing: " + self.script_filename)
            script_file.write(self._header.format(plot_filename=self.plot_filename, xlabel=xlabel, ylabel=ylabel,
                                                  title=detector.name))
            plt_cmd = self._plotting_command[detector.dimension]

            # add error plot if error data present
            err_cmd = ""
            if np.any(detector.error):
                err_cmd = self._error_plot_command.format(data_filename=self.data_filename)

            script_file.write(plt_cmd.format(data_filename=self.data_filename, error_plot=err_cmd))


class ImageWriter:
    def __init__(self, filename, options):
        self.plot_filename = filename
        if not self.plot_filename.endswith(".png"):
            self.plot_filename += ".png"
        self.colormap = options.colormap
        self.axis_with_logscale = {PlotAxis[name] for name in options.log}

    default_colormap = 'gnuplot2'

    @staticmethod
    def _make_label(unit, name):
        return name + " " + "[" + unit + "]"

    def _save_2d_error_plot(self, detector, xlist, ylist, elist, x_axis_label, y_axis_label, z_axis_label):
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        from matplotlib import colors

        # configure logscale on X and Y axis (both for positive and negative numbers)

        fig, ax = plt.subplots(1, 1)

        if PlotAxis.x in self.axis_with_logscale:
            plt.xscale('symlog')
        if PlotAxis.y in self.axis_with_logscale:
            plt.yscale('symlog')

        if PlotAxis.z in self.axis_with_logscale:
            norm = colors.LogNorm(vmin=elist[elist > 0].min(), vmax=elist.max())
        else:
            norm = colors.Normalize(vmin=elist.min(), vmax=elist.max())

        plt.xlabel(x_axis_label)
        plt.ylabel(y_axis_label)

        mesh = plt.pcolormesh(xlist, ylist, elist.clip(0.0), cmap=self.colormap, norm=norm)
        cbar = fig.colorbar(mesh)
        cbar.set_label(label=z_axis_label, rotation=270, verticalalignment='bottom')

        base_name, _ = os.path.splitext(self.plot_filename)
        plt.savefig(base_name + "_error.png")
        plt.close()

    def write(self, detector):
        try:
            import matplotlib
            matplotlib.use('Agg')
            import matplotlib.pyplot as plt
            from matplotlib import colors
        except ImportError:
            logger.error("Matplotlib not installed, output won't be generated")
            return

        # skip plotting 0-D and 3-D data
        if detector.dimension in (0, 3):
            return

        data_raw = detector.data_raw
        error_raw = detector.error_raw

        # change units for LET from MeV/cm to keV/um if necessary
        # a copy of datatable is made here
        from pymchelper.shieldhit.detector.detector_type import SHDetType
        if detector.dettyp in (SHDetType.dlet, SHDetType.dletg, SHDetType.tlet, SHDetType.tletg):
            data_raw = data_raw * np.float64(0.1)  # 1 MeV / cm = 0.1 keV / um
            if not np.all(np.isnan(error_raw)) and np.any(error_raw):
                error_raw = error_raw * np.float64(0.1)  # 1 MeV / cm = 0.1 keV / um

        logger.info("Writing: " + self.plot_filename)

        plot_x_axis = detector.plot_axis(0)

        plt.xlabel(self._make_label(plot_x_axis.unit, plot_x_axis.name))

        # configure logscale on X and Y axis (both for positive and negative numbers)
        if PlotAxis.x in self.axis_with_logscale:
            plt.xscale('symlog')

        if PlotAxis.y in self.axis_with_logscale:
            plt.yscale('symlog')

        # 1-D plotting
        if detector.dimension == 1:

            # add optional error area
            if np.any(detector.error):
                plt.fill_between(plot_x_axis.data,
                                 (data_raw - error_raw).clip(0.0),
                                 (data_raw + error_raw).clip(0.0, 1.05 * data_raw.max()),
                                 alpha=0.2, edgecolor='#CC4F1B', facecolor='#FF9848', antialiased=True)
            plt.ylabel(self._make_label(detector.unit, detector.name))
            plt.plot(plot_x_axis.data, data_raw)
        elif detector.dimension == 2:
            plot_y_axis = detector.plot_axis(1)

            xlist, ylist = np.meshgrid(plot_x_axis.data, plot_y_axis.data)

            x_axis_label = self._make_label(plot_x_axis.unit, plot_x_axis.name)
            y_axis_label = self._make_label(plot_y_axis.unit, plot_y_axis.name)
            z_axis_label = self._make_label(detector.unit, detector.name)

            # configure logscale on Z axis
            if PlotAxis.z in self.axis_with_logscale:
                norm = colors.LogNorm(vmin=data_raw[data_raw > 0].min(), vmax=data_raw.max())
            else:
                norm = colors.Normalize(vmin=data_raw.min(), vmax=data_raw.max())

            # in case differential scorer was used there is a case when axis has to be swapped
            # this happens when X-constant, Y-differential, Z-scored
            if hasattr(detector, 'dif_axis') and detector.dif_axis == 1:
                x_axis_label, y_axis_label = y_axis_label, x_axis_label
                xlist, ylist = ylist, xlist

            plt.xlabel(x_axis_label)
            plt.ylabel(y_axis_label)

            shape_tuple = (plot_y_axis.n, plot_x_axis.n)
            zlist = data_raw.reshape(shape_tuple)
            plt.pcolormesh(xlist, ylist, zlist, cmap=self.colormap, norm=norm)

            cbar = plt.colorbar()
            cbar.set_label(z_axis_label, rotation=270, verticalalignment='bottom')

        plt.savefig(self.plot_filename)
        plt.close()

        # add 2-D error plot if error data present
        if detector.dimension == 2 and not np.all(np.isnan(error_raw)) and np.any(error_raw):
            elist = error_raw.reshape(shape_tuple)
            self._save_2d_error_plot(detector, xlist, ylist, elist, x_axis_label, y_axis_label, z_axis_label)



import sys
from itertools import product

# import necessary code pieces from pymchelper
from pymchelper.shieldhit.detector.detector_type import SHDetType
from pymchelper.shieldhit.detector.estimator import SHEstimator
from pymchelper.shieldhit.detector.estimator_type import SHGeoType
from pymchelper.shieldhit.detector.fortran_card import EstimatorWriter, CardLine
from pymchelper.shieldhit.detector.geometry import CarthesianMesh
from pymchelper.shieldhit.particle import SHParticleType


def main(args=sys.argv[1:]):
    """
    Compose programatically detect.dat SHIELDHIT-12A input file
    with fixed mesh and many combinations of detector and particle type.
    :param args: part of sys.argv, used here to simplify automated testing
    :return: None
    """

    # create empty estimator object
    estimator = SHEstimator()

    # create carthesian mesh
    # it is done once and in single place
    # editing in manually in detect.dat file would require changes in many lines,
    # for every combination of particle and detector type, making it error prone
    estimator.estimator = SHGeoType.msh
    estimator.geometry = CarthesianMesh()
    estimator.geometry.set_axis(axis_no=0, start=-5.0, stop=5.0, nbins=1)
    estimator.geometry.set_axis(axis_no=1, start=-5.0, stop=5.0, nbins=1)
    estimator.geometry.set_axis(axis_no=2, start=0.0, stop=30.0, nbins=300)

    # possible detector types and associated names
    det_types = {SHDetType.energy: "en", SHDetType.fluence: "fl"}

    # possible particle types and associated names
    particle_types = {SHParticleType.all: "all", SHParticleType.proton: "p", SHParticleType.neutron: "n"}

    # open detector.dat file for writing
    with open("detect.dat", "w") as f:
        f.write(CardLine.credits + "\n")

        # loop over all combinations of detector and particle types
        # output filename will be composed from associated detector and particle names
        for dt, pt in product(det_types.keys(), particle_types.keys()):
            estimator.detector_type = dt
            estimator.particle_type = pt
            estimator.filename = det_types[dt] + "_" + particle_types[pt]
            text = EstimatorWriter.get_text(estimator, add_comment=True)
            f.write(text)
        f.write(CardLine.comment + "\n")

    # detect.dat file should have following content:
    #  * generated by pymchelper (https://github.com/DataMedSci/pymchelper) *
    # *----0---><----1---><----2---><----3---><----4---><----5---><----6--->
    # MSH             -5.0      -5.0       0.0       5.0       5.0      30.0
    #                    1         1       300         1    ENERGY      en_n
    # *----0---><----1---><----2---><----3---><----4---><----5---><----6--->
    # MSH             -5.0      -5.0       0.0       5.0       5.0      30.0
    #                    1         1       300         2    ENERGY      en_p
    # *----0---><----1---><----2---><----3---><----4---><----5---><----6--->
    # MSH             -5.0      -5.0       0.0       5.0       5.0      30.0
    #                    1         1       300        -1    ENERGY    en_all
    # *----0---><----1---><----2---><----3---><----4---><----5---><----6--->
    # MSH             -5.0      -5.0       0.0       5.0       5.0      30.0
    #                    1         1       300         1   FLUENCE      fl_n
    # *----0---><----1---><----2---><----3---><----4---><----5---><----6--->
    # MSH             -5.0      -5.0       0.0       5.0       5.0      30.0
    #                    1         1       300         2   FLUENCE      fl_p
    # *----0---><----1---><----2---><----3---><----4---><----5---><----6--->
    # MSH             -5.0      -5.0       0.0       5.0       5.0      30.0
    #                    1         1       300        -1   FLUENCE    fl_all
    # *----0---><----1---><----2---><----3---><----4---><----5---><----6--->


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))
